/*
 *  Copyright (C) 2016, Burt P.,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without modification,
 *  are permitted provided that the following conditions are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The names of its contributors may not be used to endorse or promote
 *       products derived from this software without specific prior written
 *       permission.
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This file exists to remove a dependency on the math lib in libhdcd
 * that only exists so that sin() can be used in the analyze mode tone
 * generator. This version uses simple lookup tables instead.
 *
 * To generate tables:
 * gcc -o tone_tab -DHDCD_PRINT_TONE_TAB hdcd_analyze_tonegen.c -lm
 * ./tone_tab >tone_tab.csnip
 */

#ifdef HDCD_PRINT_TONE_TAB
#include <stdio.h>
#include <math.h>
#endif
#include <stdint.h>

static const int16_t tone_tab_96000[320] = {
    0x0000, 0x0040, 0x0080, 0x00c0, 0x0101, 0x0141, 0x0181, 0x01c0,
    0x0200, 0x0240, 0x027f, 0x02be, 0x02fc, 0x033b, 0x0379, 0x03b7,
    0x03f4, 0x0431, 0x046e, 0x04aa, 0x04e5, 0x0521, 0x055b, 0x0595,
    0x05cf, 0x0608, 0x0641, 0x0678, 0x06b0, 0x06e6, 0x071c, 0x0751,
    0x0785, 0x07b9, 0x07ec, 0x081e, 0x0850, 0x0880, 0x08b0, 0x08df,
    0x090c, 0x093a, 0x0966, 0x0991, 0x09bb, 0x09e4, 0x0a0d, 0x0a34,
    0x0a5a, 0x0a80, 0x0aa4, 0x0ac7, 0x0ae9, 0x0b0a, 0x0b2a, 0x0b49,
    0x0b67, 0x0b84, 0x0b9f, 0x0bba, 0x0bd3, 0x0beb, 0x0c02, 0x0c17,
    0x0c2c, 0x0c3f, 0x0c51, 0x0c62, 0x0c72, 0x0c80, 0x0c8d, 0x0c99,
    0x0ca4, 0x0cad, 0x0cb5, 0x0cbc, 0x0cc2, 0x0cc7, 0x0cca, 0x0ccc,
    0x0ccc, 0x0ccc, 0x0cca, 0x0cc7, 0x0cc2, 0x0cbc, 0x0cb5, 0x0cad,
    0x0ca4, 0x0c99, 0x0c8d, 0x0c80, 0x0c72, 0x0c62, 0x0c51, 0x0c3f,
    0x0c2c, 0x0c17, 0x0c02, 0x0beb, 0x0bd3, 0x0bba, 0x0b9f, 0x0b84,
    0x0b67, 0x0b49, 0x0b2a, 0x0b0a, 0x0ae9, 0x0ac7, 0x0aa4, 0x0a80,
    0x0a5a, 0x0a34, 0x0a0d, 0x09e4, 0x09bb, 0x0991, 0x0966, 0x093a,
    0x090c, 0x08df, 0x08b0, 0x0880, 0x0850, 0x081e, 0x07ec, 0x07b9,
    0x0785, 0x0751, 0x071c, 0x06e6, 0x06b0, 0x0678, 0x0641, 0x0608,
    0x05cf, 0x0595, 0x055b, 0x0521, 0x04e5, 0x04aa, 0x046e, 0x0431,
    0x03f4, 0x03b7, 0x0379, 0x033b, 0x02fc, 0x02be, 0x027f, 0x0240,
    0x0200, 0x01c0, 0x0181, 0x0141, 0x0101, 0x00c0, 0x0080, 0x0040,
    0x0000, 0xffc0, 0xff80, 0xff40, 0xfeff, 0xfebf, 0xfe7f, 0xfe40,
    0xfe00, 0xfdc0, 0xfd81, 0xfd42, 0xfd04, 0xfcc5, 0xfc87, 0xfc49,
    0xfc0c, 0xfbcf, 0xfb92, 0xfb56, 0xfb1b, 0xfadf, 0xfaa5, 0xfa6b,
    0xfa31, 0xf9f8, 0xf9bf, 0xf988, 0xf950, 0xf91a, 0xf8e4, 0xf8af,
    0xf87b, 0xf847, 0xf814, 0xf7e2, 0xf7b0, 0xf780, 0xf750, 0xf721,
    0xf6f4, 0xf6c6, 0xf69a, 0xf66f, 0xf645, 0xf61c, 0xf5f3, 0xf5cc,
    0xf5a6, 0xf580, 0xf55c, 0xf539, 0xf517, 0xf4f6, 0xf4d6, 0xf4b7,
    0xf499, 0xf47c, 0xf461, 0xf446, 0xf42d, 0xf415, 0xf3fe, 0xf3e9,
    0xf3d4, 0xf3c1, 0xf3af, 0xf39e, 0xf38e, 0xf380, 0xf373, 0xf367,
    0xf35c, 0xf353, 0xf34b, 0xf344, 0xf33e, 0xf339, 0xf336, 0xf334,
    0xf334, 0xf334, 0xf336, 0xf339, 0xf33e, 0xf344, 0xf34b, 0xf353,
    0xf35c, 0xf367, 0xf373, 0xf380, 0xf38e, 0xf39e, 0xf3af, 0xf3c1,
    0xf3d4, 0xf3e9, 0xf3fe, 0xf415, 0xf42d, 0xf446, 0xf461, 0xf47c,
    0xf499, 0xf4b7, 0xf4d6, 0xf4f6, 0xf517, 0xf539, 0xf55c, 0xf580,
    0xf5a6, 0xf5cc, 0xf5f3, 0xf61c, 0xf645, 0xf66f, 0xf69a, 0xf6c6,
    0xf6f4, 0xf721, 0xf750, 0xf780, 0xf7b0, 0xf7e2, 0xf814, 0xf847,
    0xf87b, 0xf8af, 0xf8e4, 0xf91a, 0xf950, 0xf988, 0xf9bf, 0xf9f8,
    0xfa31, 0xfa6b, 0xfaa5, 0xfadf, 0xfb1b, 0xfb56, 0xfb92, 0xfbcf,
    0xfc0c, 0xfc49, 0xfc87, 0xfcc5, 0xfd04, 0xfd42, 0xfd81, 0xfdc0,
    0xfe00, 0xfe40, 0xfe7f, 0xfebf, 0xfeff, 0xff40, 0xff80, 0xffc0,

};

static const int16_t tone_tab_88200[294] = {
    0x0000, 0x0046, 0x008c, 0x00d1, 0x0117, 0x015d, 0x01a3, 0x01e8,
    0x022d, 0x0272, 0x02b6, 0x02fb, 0x033f, 0x0382, 0x03c5, 0x0408,
    0x044a, 0x048c, 0x04cd, 0x050e, 0x054e, 0x058d, 0x05cc, 0x060a,
    0x0647, 0x0684, 0x06c0, 0x06fb, 0x0735, 0x076f, 0x07a7, 0x07df,
    0x0816, 0x084c, 0x0880, 0x08b4, 0x08e7, 0x0919, 0x094a, 0x0979,
    0x09a8, 0x09d5, 0x0a01, 0x0a2c, 0x0a56, 0x0a7f, 0x0aa6, 0x0acd,
    0x0af2, 0x0b15, 0x0b38, 0x0b59, 0x0b78, 0x0b97, 0x0bb4, 0x0bcf,
    0x0bea, 0x0c03, 0x0c1a, 0x0c30, 0x0c45, 0x0c58, 0x0c6a, 0x0c7a,
    0x0c89, 0x0c96, 0x0ca2, 0x0cad, 0x0cb6, 0x0cbd, 0x0cc3, 0x0cc8,
    0x0ccb, 0x0ccc, 0x0ccc, 0x0ccb, 0x0cc8, 0x0cc3, 0x0cbd, 0x0cb6,
    0x0cad, 0x0ca2, 0x0c96, 0x0c89, 0x0c7a, 0x0c6a, 0x0c58, 0x0c45,
    0x0c30, 0x0c1a, 0x0c03, 0x0bea, 0x0bcf, 0x0bb4, 0x0b97, 0x0b78,
    0x0b59, 0x0b38, 0x0b15, 0x0af2, 0x0acd, 0x0aa6, 0x0a7f, 0x0a56,
    0x0a2c, 0x0a01, 0x09d5, 0x09a8, 0x0979, 0x094a, 0x0919, 0x08e7,
    0x08b4, 0x0880, 0x084c, 0x0816, 0x07df, 0x07a7, 0x076f, 0x0735,
    0x06fb, 0x06c0, 0x0684, 0x0647, 0x060a, 0x05cc, 0x058d, 0x054e,
    0x050e, 0x04cd, 0x048c, 0x044a, 0x0408, 0x03c5, 0x0382, 0x033f,
    0x02fb, 0x02b6, 0x0272, 0x022d, 0x01e8, 0x01a3, 0x015d, 0x0117,
    0x00d1, 0x008c, 0x0046, 0x0000, 0xffba, 0xff74, 0xff2f, 0xfee9,
    0xfea3, 0xfe5d, 0xfe18, 0xfdd3, 0xfd8e, 0xfd4a, 0xfd05, 0xfcc1,
    0xfc7e, 0xfc3b, 0xfbf8, 0xfbb6, 0xfb74, 0xfb33, 0xfaf2, 0xfab2,
    0xfa73, 0xfa34, 0xf9f6, 0xf9b9, 0xf97c, 0xf940, 0xf905, 0xf8cb,
    0xf891, 0xf859, 0xf821, 0xf7ea, 0xf7b4, 0xf780, 0xf74c, 0xf719,
    0xf6e7, 0xf6b6, 0xf687, 0xf658, 0xf62b, 0xf5ff, 0xf5d4, 0xf5aa,
    0xf581, 0xf55a, 0xf533, 0xf50e, 0xf4eb, 0xf4c8, 0xf4a7, 0xf488,
    0xf469, 0xf44c, 0xf431, 0xf416, 0xf3fd, 0xf3e6, 0xf3d0, 0xf3bb,
    0xf3a8, 0xf396, 0xf386, 0xf377, 0xf36a, 0xf35e, 0xf353, 0xf34a,
    0xf343, 0xf33d, 0xf338, 0xf335, 0xf334, 0xf334, 0xf335, 0xf338,
    0xf33d, 0xf343, 0xf34a, 0xf353, 0xf35e, 0xf36a, 0xf377, 0xf386,
    0xf396, 0xf3a8, 0xf3bb, 0xf3d0, 0xf3e6, 0xf3fd, 0xf416, 0xf431,
    0xf44c, 0xf469, 0xf488, 0xf4a7, 0xf4c8, 0xf4eb, 0xf50e, 0xf533,
    0xf55a, 0xf581, 0xf5aa, 0xf5d4, 0xf5ff, 0xf62b, 0xf658, 0xf687,
    0xf6b6, 0xf6e7, 0xf719, 0xf74c, 0xf780, 0xf7b4, 0xf7ea, 0xf821,
    0xf859, 0xf891, 0xf8cb, 0xf905, 0xf940, 0xf97c, 0xf9b9, 0xf9f6,
    0xfa34, 0xfa73, 0xfab2, 0xfaf2, 0xfb33, 0xfb74, 0xfbb6, 0xfbf8,
    0xfc3b, 0xfc7e, 0xfcc1, 0xfd05, 0xfd4a, 0xfd8e, 0xfdd3, 0xfe18,
    0xfe5d, 0xfea3, 0xfee9, 0xff2f, 0xff74, 0xffba,
};

/* prototype to avoid a warning. */
int _hdcd_tone16(int *sn, int rate);

int _hdcd_tone16(int *sn, int rate)
{
    (*sn)++;
    switch(rate) {
        case 96000:
            *sn %= 320;
            return tone_tab_96000[*sn];
        case 48000:
            *sn %= 160;
            return tone_tab_96000[(*sn)<<1];
        case 88200:
            *sn %= 294;
            return tone_tab_88200[*sn];
        case 44100:
        default:
            *sn %= 147;
            return tone_tab_88200[(*sn)<<1];
    }
}

#ifdef HDCD_PRINT_TONE_TAB

/** tone generator: sample_number, frequency, sample_rate, amplitude */
#define TONEGEN16(sn, f, sr, a) (int16_t)(sin((6.28318530718 * (sn) * (f)) /(sr)) * (a) * 0x7fff)

int main()
{
    int n;
    int freq = 300;
    int rate = 96000;
    int max = rate / freq;

    printf("static const int16_t tone_tab_%d[%d] = {\n    ", rate, max);
    for (n = 0; n < max; n++) {
        printf("0x%04hx, ", TONEGEN16(n, freq, rate, 0.1) );
        if (n%8 == 7) printf("\n    ");
    }
    printf("\n};\n");
    rate = 88200;
    max = rate / freq;
    printf("static const int16_t tone_tab_%d[%d] = {\n    ", rate, max);
    for (n = 0; n < max; n++) {
        printf("0x%04hx, ", TONEGEN16(n, freq, rate, 0.1) );
        if (n%8 == 7) printf("\n    ");
    }
    printf("\n};\n");
}
#endif
